@@ -3,9 +3,11 @@ use strict;
 use warnings;
 use Carp;
 use Data::Dumper;
-our $VERSION = '2.1.1';
+use Scalar::Util qw( weaken );
+our $VERSION = '2.1.2';
 
 my $GlobalRef = {};
+
 my $THIS;
 my $DUKTAPE;
 my $isNew  = bless [], "NEW";
@@ -142,8 +144,6 @@ sub new {
 
     $THIS = bless { duk => $duk, heapptr => 0 }, "JavaScript::Duktape::Object";
 
-    ##global methods
-
     ##finalizer method
     $self->{finalizer} = sub {
         my $ref = $duk->get_string(0);
@@ -151,36 +151,11 @@ sub new {
         return 1;
     };
 
+    weaken $GlobalRef;
+
     $duk->perl_push_function($self->{finalizer}, 1);
     $duk->put_global_string('perlFinalizer');
 
-    ##calling cached functions
-    $self->{call} = sub {
-        my $heapptr = $duk->get_heapptr(0);
-        my $name    = $duk->get_string(1);
-        my $top     = $duk->get_top();
-
-        shift if ref $_[0] eq 'JavaScript::Duktape::Cache';
-
-        $THIS->{heapptr} = $heapptr;
-        $THIS->{duk}     = $duk;
-
-        my @args;
-        if (!@_){
-            for (my $i = 2; $i < $top; $i++){
-                push @args, $duk->to_perl($i);
-            }
-        } else {
-            @args = @_;
-        }
-
-        my $ret = $GlobalRef->{$name}->{sub}->(@args);
-        $duk->push_perl($ret);
-        return 1;
-    };
-
-    $duk->push_perl_function($self->{call}, -1);
-    $duk->put_global_string('perlCall');
     return $self;
 }
 
@@ -229,7 +204,7 @@ sub get {
     my $duk = $self->vm;
     $duk->push_string($name);
     if ($duk->peval() != 0) {
-        croak $duk->safe_to_string(-1);
+        croak $duk->last_error_string();
     }
     my $ret = $duk->to_perl(-1);
     $duk->pop();
@@ -242,7 +217,7 @@ sub get_object {
     my $duk = $self->vm;
     $duk->push_string($name);
     if ($duk->peval() != 0) {
-        croak $duk->safe_to_string(-1);
+        croak $duk->last_error_string();
     }
     my $ret = $duk->to_perl_object(-1);
     $duk->pop();
@@ -255,11 +230,8 @@ sub eval {
     my $string = shift;
     my $duk = $self->duk;
 
-    my $err = $duk->peval_string($string);
-
-    if ($err){
-        my $error_string = $duk->safe_to_string(-1);
-        croak $error_string;
+    if ($duk->peval_string($string) != 0){
+        croak $duk->last_error_string();
     }
 
     return $duk->to_perl(-1);
@@ -269,9 +241,9 @@ sub vm  { shift->{duk}; }
 sub duk { shift->{duk}; }
 
 sub destroy {
+    local $@;
     my $self = shift;
     my $duk  = delete $self->{duk};
-    local $@; #duk_desatroy_heap mess with $@!!
     return if !$duk;
     $duk->destroy_heap();
 }
@@ -447,7 +419,7 @@ sub to_perl {
                     $self->push_perl($_[$i]);
                 }
                 if ($self->pcall_method($len) == 1) {
-                    croak $self->safe_to_string(-1);
+                    croak $self->last_error_string();
                 }
                 my $ret = $self->to_perl(-1);
                 $self->pop();
@@ -548,7 +520,9 @@ sub push_function {
     }, $nargs);
 }
 
-
+#####################################################################
+# safe call
+#####################################################################
 sub push_c_function {
     my $self  = shift;
     my $sub   = shift;
@@ -558,31 +532,15 @@ sub push_c_function {
         my @args = @_;
         my $top = $self->get_top();
         my $ret = 1;
-        my $died;
-        $self->perl_duk_safe_call(sub {
-            eval { $ret = $sub->(@args) };
-            my $error = $@;
-            if ($error){
-                if ($error =~ /^Duk::Error/){
-                    croak $@;
-                } else {
-                    ## don't throw error inside perl sub safe call
-                    ## just tell our sub that there is an error and
-                    ## let it exit this calling stack then throw
-                    ## Fixes issue #6
-                    $died = $error;
-                    $self->eval_string('(function (e){ return new Error(e) })');
-                    $self->push_string($error);
-                    $self->call(1);
-                }
-            }
-            return $ret;
+
+        my $err = $self->safe_call( sub {
+            $ret = $sub->(@args);
+            return 1;
         }, $top, 1);
 
-        if ($died){
-            croak $died;
+        if ($err) {
+            croak $self->last_error_string();
         }
-
         return $ret;
     };
 
@@ -592,58 +550,32 @@ sub push_c_function {
 }
 
 
-sub cache {
-    my $self = shift;
-    my $sub = shift;
-
-    my @caller = caller;
-    my $code_cache_name = $caller[0] . $caller[2];
-
-    if (!$GlobalRef->{$code_cache_name}){
-        $GlobalRef->{$code_cache_name} = bless {
-            duk => $self,
-            sub => $sub,
-            name => $code_cache_name
-        }, "JavaScript::Duktape::Cache";
-    }
-    $GlobalRef->{$code_cache_name}->{args} = \@_;
-    return $GlobalRef->{$code_cache_name};
-}
-
-
 #####################################################################
 # safe call
 #####################################################################
 sub safe_call {
     my $self = shift;
     my $sub = shift;
+    my $ret;
     my $safe = sub {
-        return $sub->($self);
+        local $@;
+        eval { $ret = $sub->($self) };
+        if ( my $error = $@ ) {
+            if ($error =~ /^Duk::Error/i) {
+                croak $self->last_error_string();
+            }
+            else {
+                $self->eval_string('(function (e){ throw new Error(e) })');
+                $self->push_string($error);
+                $self->call(1);
+            }
+        }
+
+        return defined $ret ? $ret : 1;
     };
 
-    my $oldtop = $self->get_top();
-    eval { $self->perl_duk_safe_call($safe, @_) };
-    my $error = $@;
-    if ($error) {
-        if ($error =~ /^Duk::Error/i) {
-            # error came from duktape do nothing;
-        } else {
-            #error from perl push error object
-            #and clear previous pushed stack
-            my $newtop = $self->get_top();
-            if ($newtop > $oldtop){
-                $self->pop_n($newtop - $oldtop);
-            }
-
-            $self->eval_string('(function (e){ return new Error(e) })');
-            $self->push_string($error);
-            $self->call(1);
-        }
-        return 1;
-    }
-
-    #safe_call success returns 0
-    return 0;
+    eval { $ret = $self->perl_duk_safe_call($safe, @_) };
+    return defined $ret ? $ret : 1;
 }
 
 
@@ -666,6 +598,13 @@ sub safe_call {
 ##############################################
 *reset_top = \&perl_duk_reset_top;
 
+sub last_error_string {
+    my $self = shift;
+    $self->dup(-1);
+    my $error_str =  $self->safe_to_string(-1);
+    $self->pop();
+    return $error_str;
+}
 
 sub dump {
     my $self = shift;
@@ -814,7 +753,7 @@ package JavaScript::Duktape::Util; {
                 $duk->push_perl($val);
             }
             if ($duk->pnew($len) != 0){
-                croak $duk->safe_to_string(-1);
+                croak $duk->last_error_string();
             }
             my $val = $duk->to_perl_object(-1);
             $duk->pop();
@@ -874,14 +813,6 @@ package JavaScript::Duktape::Util; {
             foreach my $val (@_){
                 if (ref $val eq 'CODE'){
                     $duk->push_function($val);
-                } elsif (ref $val eq 'JavaScript::Duktape::Cache') {
-                    my $name = $val->{name};
-                    $duk->eval_string(qq~
-                        (function (a, b, c, d){
-                            var name = "$name";
-                            return perlCall(this, name, a, b, c, d);
-                        });
-                    ~);
                 } else {
                     $duk->push_perl($val);
                 }
@@ -889,11 +820,11 @@ package JavaScript::Duktape::Util; {
 
             if ($isNew){
                 if ($duk->pnew($len) != 0){
-                    croak $duk->safe_to_string(-1);
+                    croak $duk->last_error_string();
                 }
             } else {
                 if ($duk->pcall_method($len) != 0){
-                    croak $duk->safe_to_string(-1);
+                    croak $duk->last_error_string();
                 }
             }
 
